"""Distance calculation of specific geometric shapes.

Note that for distance calculations to triangles, rectangles, boxes, cylinders,
spheres, capsules, and convex meshes GJK is usually the fastest option if
applicable. GJK is not applicable if one of the geometric objects is a line
or a plane and GJK is slower when one of the objects is a point or a line
segment.
"""
from ._line import (
    point_to_line, line_to_line, point_to_line_segment, line_to_line_segment,
    line_segment_to_line_segment)
from ._plane import point_to_plane, line_to_plane, line_segment_to_plane
from ._triangle import (
    point_to_triangle, line_to_triangle, line_segment_to_triangle,
    triangle_to_triangle, triangle_to_rectangle)
from ._rectangle import (
    point_to_rectangle, line_to_rectangle, line_segment_to_rectangle,
    rectangle_to_rectangle)
from ._disk import point_to_disk, disk_to_disk
from ._circle import point_to_circle, line_to_circle
from ._box import (
    point_to_box, line_to_box, line_segment_to_box, rectangle_to_box)
from ._ellipsoid import point_to_ellipsoid
from ._cylinder import point_to_cylinder


__all__ = [
    "point_to_line",
    "point_to_line_segment",
    "point_to_plane",
    "point_to_triangle",
    "point_to_rectangle",
    "point_to_disk",
    "point_to_circle",
    "point_to_box",
    "point_to_ellipsoid",
    "point_to_cylinder",
    "line_to_line",
    "line_to_line_segment",
    "line_to_plane",
    "line_to_triangle",
    "line_to_rectangle",
    "line_to_circle",
    "line_to_box",
    "line_segment_to_line_segment",
    "line_segment_to_plane",
    "line_segment_to_triangle",
    "line_segment_to_rectangle",
    "line_segment_to_box",
    "triangle_to_triangle",
    "triangle_to_rectangle",
    "rectangle_to_rectangle",
    "rectangle_to_box",
    "disk_to_disk",
]



"""Distance between line and box.

Implementation adapted from 3D Game Engine Design by David H. Eberly.

Geometric Tools, Inc.
http://www.geometrictools.com
Copyright (c) 1998-2006.  All Rights Reserved

The Wild Magic Version 4 Foundation Library source code is supplied
under the terms of the license agreement
(http://www.geometrictools.com/License/Wm4FoundationLicense.pdf)
and may not be copied or disclosed except in accordance with the terms
of that agreement.
"""
import math

import numpy as np


def _line_to_box(line_point, line_direction, box2origin, size, origin2box=None):
    box_half_size = 0.5 * size

    # compute coordinates of line in box coordinate system
    if origin2box is None:
        origin2box = np.linalg.inv(box2origin)
    point_in_box = origin2box[:3, 3] + origin2box[:3, :3].dot(line_point)
    direction_in_box = origin2box[:3, :3].dot(line_direction)

    # Apply reflections so that direction vector has nonnegative components.
    direction_sign = np.ones(3)
    for i in range(3):
        if direction_in_box[i] < 0.0:
            point_in_box[i] = -point_in_box[i]
            direction_in_box[i] = -direction_in_box[i]
            direction_sign[i] = -1.0

    if direction_in_box[0] > 0.0:
        if direction_in_box[1] > 0.0:
            if direction_in_box[2] > 0.0:  # (+,+,+)
                sqr_dist, line_parameter = _case_no_zeros(
                    point_in_box, direction_in_box, box_half_size)
            else:  # (+,+,0)
                sqr_dist, line_parameter = _case_0(
                    0, 1, 2, point_in_box, direction_in_box, box_half_size)
        else:
            if direction_in_box[2] > 0.0:  # (+,0,+)
                sqr_dist, line_parameter = _case_0(
                    0, 2, 1, point_in_box, direction_in_box, box_half_size)
            else:  # (+,0,0)
                sqr_dist, line_parameter = _case_00(
                    0, 1, 2, point_in_box, direction_in_box, box_half_size)
    else:
        if direction_in_box[1] > 0.0:
            if direction_in_box[2] > 0.0:  # (0,+,+)
                sqr_dist, line_parameter = _case_0(
                    1, 2, 0, point_in_box, direction_in_box, box_half_size)
            else:  # (0,+,0)
                sqr_dist, line_parameter = _case_00(
                    1, 0, 2, point_in_box, direction_in_box, box_half_size)
        else:
            if direction_in_box[2] > 0.0:  # (0,0,+)
                sqr_dist, line_parameter = _case_00(
                    2, 0, 1, point_in_box, direction_in_box, box_half_size)
            else:  # (0,0,0)
                sqr_dist = _case_000(point_in_box, box_half_size)
                line_parameter = 0.0

    # compute closest point on line
    contact_point_line = line_point + line_parameter * line_direction

    # compute closest point on box
    # undo the reflections applied previously
    contact_point_box = box2origin[:3, 3] + box2origin[:3, :3].dot(
        direction_sign * point_in_box)

    return math.sqrt(sqr_dist), contact_point_line, contact_point_box, line_parameter


def _case_no_zeros(point_in_box, direction_in_box, box_half_size):
    kPmE = point_in_box - box_half_size

    fProdDxPy = direction_in_box[0] * kPmE[1]
    fProdDyPx = direction_in_box[1] * kPmE[0]

    if fProdDyPx >= fProdDxPy:
        fProdDzPx = direction_in_box[2] * kPmE[0]
        fProdDxPz = direction_in_box[0] * kPmE[2]
        if fProdDzPx >= fProdDxPz:
            # line intersects x = e0
            sqr_dist, line_parameter = _box_face(0, 1, 2, point_in_box, direction_in_box, kPmE, box_half_size)
        else:
            # line intersects z = e2
            sqr_dist, line_parameter = _box_face(2, 0, 1, point_in_box, direction_in_box, kPmE, box_half_size)
    else:
        fProdDzPy = direction_in_box[2] * kPmE[1]
        fProdDyPz = direction_in_box[1] * kPmE[2]
        if fProdDzPy >= fProdDyPz:
            # line intersects y = e1
            sqr_dist, line_parameter = _box_face(1, 2, 0, point_in_box, direction_in_box, kPmE, box_half_size)
        else:
            # line intersects z = e2
            sqr_dist, line_parameter = _box_face(2, 0, 1, point_in_box, direction_in_box, kPmE, box_half_size)
    return sqr_dist, line_parameter


def _box_face(i0, i1, i2, point_in_box, direction_in_box, rkPmE, box_half_size):
    sqr_dist = 0.0
    kPpE = np.zeros(3)

    kPpE[i1] = point_in_box[i1] + box_half_size[i1]
    kPpE[i2] = point_in_box[i2] + box_half_size[i2]
    if direction_in_box[i0] * kPpE[i1] >= direction_in_box[i1] * rkPmE[i0]:
        if direction_in_box[i0] * kPpE[i2] >= direction_in_box[i2] * rkPmE[i0]:
            # v[i1] >= -e[i1], v[i2] >= -e[i2] (distance = 0)
            point_in_box[i0] = box_half_size[i0]
            point_in_box[i1] -= direction_in_box[i1] * rkPmE[i0] / direction_in_box[i0]
            point_in_box[i2] -= direction_in_box[i2] * rkPmE[i0] / direction_in_box[i0]
            line_parameter = -rkPmE[i0] / direction_in_box[i0]
        else:
            # v[i1] >= -e[i1], v[i2] < -e[i2]
            fLSqr = direction_in_box[i0] * direction_in_box[i0] + direction_in_box[i2] * direction_in_box[i2]
            fTmp = fLSqr * kPpE[i1] - direction_in_box[i1] * (direction_in_box[i0] * rkPmE[i0] + direction_in_box[i2] * kPpE[i2])
            if fTmp <= 2.0 * fLSqr * box_half_size[i1]:
                fT = fTmp / fLSqr
                fLSqr += direction_in_box[i1] * direction_in_box[i1]
                fTmp = kPpE[i1] - fT
                delta = direction_in_box[i0] * rkPmE[i0] + direction_in_box[i1] * fTmp + direction_in_box[i2] * kPpE[i2]
                line_parameter = -delta / fLSqr
                sqr_dist += rkPmE[i0] * rkPmE[i0] + fTmp * fTmp + kPpE[i2] * kPpE[i2] + delta * line_parameter

                point_in_box[i0] = box_half_size[i0]
                point_in_box[i1] = fT - box_half_size[i1]
                point_in_box[i2] = -box_half_size[i2]
            else:
                fLSqr += direction_in_box[i1] * direction_in_box[i1]
                delta = direction_in_box[i0] * rkPmE[i0] + direction_in_box[i1] * rkPmE[i1] + direction_in_box[i2] * kPpE[i2]
                line_parameter = -delta / fLSqr
                sqr_dist += rkPmE[i0] * rkPmE[i0] + rkPmE[i1] * rkPmE[i1] + kPpE[i2] * kPpE[i2] + delta * line_parameter

                point_in_box[i0] = box_half_size[i0]
                point_in_box[i1] = box_half_size[i1]
                point_in_box[i2] = -box_half_size[i2]
    else:
        if direction_in_box[i0] * kPpE[i2] >= direction_in_box[i2] * rkPmE[i0]:
            # v[i1] < -e[i1], v[i2] >= -e[i2]
            fLSqr = direction_in_box[i0] * direction_in_box[i0] + direction_in_box[i1] * direction_in_box[i1]
            fTmp = fLSqr * kPpE[i2] - direction_in_box[i2] * (direction_in_box[i0] * rkPmE[i0] + direction_in_box[i1] * kPpE[i1])
            if fTmp <= 2.0 * fLSqr * box_half_size[i2]:
                fT = fTmp / fLSqr
                fLSqr += direction_in_box[i2] * direction_in_box[i2]
                fTmp = kPpE[i2] - fT
                delta = direction_in_box[i0] * rkPmE[i0] + direction_in_box[i1] * kPpE[i1] + direction_in_box[i2] * fTmp
                line_parameter = -delta / fLSqr
                sqr_dist += rkPmE[i0] * rkPmE[i0] + kPpE[i1] * kPpE[i1] + fTmp * fTmp + delta * line_parameter

                point_in_box[i0] = box_half_size[i0]
                point_in_box[i1] = -box_half_size[i1]
                point_in_box[i2] = fT - box_half_size[i2]
            else:
                fLSqr += direction_in_box[i2] * direction_in_box[i2]
                delta = direction_in_box[i0] * rkPmE[i0] + direction_in_box[i1] * kPpE[i1] + direction_in_box[i2] * rkPmE[i2]
                line_parameter = -delta / fLSqr
                sqr_dist += rkPmE[i0] * rkPmE[i0] + kPpE[i1] * kPpE[i1] + rkPmE[i2] * rkPmE[i2] + delta * line_parameter

                point_in_box[i0] = box_half_size[i0]
                point_in_box[i1] = -box_half_size[i1]
                point_in_box[i2] = box_half_size[i2]
        else:
            # v[i1] < -e[i1], v[i2] < -e[i2]
            fLSqr = direction_in_box[i0] * direction_in_box[i0] + direction_in_box[i2] * direction_in_box[i2]
            fTmp = fLSqr * kPpE[i1] - direction_in_box[i1] * (direction_in_box[i0] * rkPmE[i0] + direction_in_box[i2] * kPpE[i2])
            if fTmp >= 0.0:
                # v[i1]-edge is closest
                if fTmp <= 2.0 * fLSqr * box_half_size[i1]:
                    fT = fTmp / fLSqr
                    fLSqr += direction_in_box[i1] * direction_in_box[i1]
                    fTmp = kPpE[i1] - fT
                    delta = direction_in_box[i0] * rkPmE[i0] + direction_in_box[i1] * fTmp + direction_in_box[i2] * kPpE[i2]
                    line_parameter = -delta / fLSqr
                    sqr_dist += rkPmE[i0] * rkPmE[i0] + fTmp * fTmp + kPpE[i2] * kPpE[i2] + delta * line_parameter

                    point_in_box[i0] = box_half_size[i0]
                    point_in_box[i1] = fT - box_half_size[i1]
                    point_in_box[i2] = -box_half_size[i2]
                else:
                    fLSqr += direction_in_box[i1] * direction_in_box[i1]
                    delta = direction_in_box[i0] * rkPmE[i0] + direction_in_box[i1] * rkPmE[i1] + direction_in_box[i2] * kPpE[i2]
                    line_parameter = -delta / fLSqr
                    sqr_dist += rkPmE[i0] * rkPmE[i0] + rkPmE[i1] * rkPmE[i1] + kPpE[i2] * kPpE[
                        i2] + delta * line_parameter

                    point_in_box[i0] = box_half_size[i0]
                    point_in_box[i1] = box_half_size[i1]
                    point_in_box[i2] = -box_half_size[i2]
            else:
                fLSqr = direction_in_box[i0] * direction_in_box[i0] + direction_in_box[i1] * direction_in_box[i1]
                fTmp = fLSqr * kPpE[i2] - direction_in_box[i2] * (direction_in_box[i0] * rkPmE[i0] + direction_in_box[i1] * kPpE[i1])
                if fTmp >= 0.0:
                    # v[i2]-edge is closest
                    if fTmp <= 2.0 * fLSqr * box_half_size[i2]:
                        fT = fTmp / fLSqr
                        fLSqr += direction_in_box[i2] * direction_in_box[i2]
                        fTmp = kPpE[i2] - fT
                        delta = direction_in_box[i0] * rkPmE[i0] + direction_in_box[i1] * kPpE[i1] + direction_in_box[i2] * fTmp
                        line_parameter = -delta / fLSqr
                        sqr_dist += rkPmE[i0] * rkPmE[i0] + kPpE[i1] * kPpE[i1] + fTmp * fTmp + delta * line_parameter

                        point_in_box[i0] = box_half_size[i0]
                        point_in_box[i1] = -box_half_size[i1]
                        point_in_box[i2] = fT - box_half_size[i2]
                    else:
                        fLSqr += direction_in_box[i2] * direction_in_box[i2]
                        delta = direction_in_box[i0] * rkPmE[i0] + direction_in_box[i1] * kPpE[i1] + direction_in_box[i2] * rkPmE[i2]
                        line_parameter = -delta / fLSqr
                        sqr_dist += rkPmE[i0] * rkPmE[i0] + kPpE[i1] * kPpE[i1] + rkPmE[i2] * rkPmE[
                            i2] + delta * line_parameter

                        point_in_box[i0] = box_half_size[i0]
                        point_in_box[i1] = -box_half_size[i1]
                        point_in_box[i2] = box_half_size[i2]
                else:
                    # (v[i1],v[i2])-corner is closest
                    fLSqr += direction_in_box[i2] * direction_in_box[i2]
                    delta = direction_in_box[i0] * rkPmE[i0] + direction_in_box[i1] * kPpE[i1] + direction_in_box[i2] * kPpE[i2]
                    line_parameter = -delta / fLSqr
                    sqr_dist += rkPmE[i0] * rkPmE[i0] + kPpE[i1] * kPpE[i1] + kPpE[i2] * kPpE[i2] + delta * line_parameter

                    point_in_box[i0] = box_half_size[i0]
                    point_in_box[i1] = -box_half_size[i1]
                    point_in_box[i2] = -box_half_size[i2]

    return sqr_dist, line_parameter


def _case_0(i0, i1, i2, point_in_box, direction_in_box, box_half_size):
    sqr_dist = 0.0
    fPmE0 = point_in_box[i0] - box_half_size[i0]
    fPmE1 = point_in_box[i1] - box_half_size[i1]
    fProd0 = direction_in_box[i1] * fPmE0
    fProd1 = direction_in_box[i0] * fPmE1

    if fProd0 >= fProd1:
        # line intersects P[i0] = e[i0]
        point_in_box[i0] = box_half_size[i0]

        fPpE1 = point_in_box[i1] + box_half_size[i1]
        delta = fProd0 - direction_in_box[i0] * fPpE1
        if delta >= 0.0:
            fInvLSqr = 1.0 / (direction_in_box[i0] * direction_in_box[i0] + direction_in_box[i1] * direction_in_box[i1])
            sqr_dist += delta * delta * fInvLSqr
            point_in_box[i1] = -box_half_size[i1]
            line_parameter = -(direction_in_box[i0] * fPmE0 + direction_in_box[i1] * fPpE1) * fInvLSqr
        else:
            fInv = 1.0 / direction_in_box[i0]
            point_in_box[i1] -= fProd0 * fInv
            line_parameter = -fPmE0 * fInv
    else:
        # line intersects P[i1] = e[i1]
        point_in_box[i1] = box_half_size[i1]

        fPpE0 = point_in_box[i0] + box_half_size[i0]
        delta = fProd1 - direction_in_box[i1] * fPpE0
        if delta >= 0.0:
            fInvLSqr = 1.0 / (direction_in_box[i0] * direction_in_box[i0] + direction_in_box[i1] * direction_in_box[i1])
            sqr_dist += delta * delta * fInvLSqr
            point_in_box[i0] = -box_half_size[i0]
            line_parameter = -(direction_in_box[i0] * fPpE0 + direction_in_box[i1] * fPmE1) * fInvLSqr
        else:
            fInv = 1.0 / direction_in_box[i1]
            point_in_box[i0] -= fProd1 * fInv
            line_parameter = -fPmE1 * fInv

    if point_in_box[i2] < -box_half_size[i2]:
        delta = point_in_box[i2] + box_half_size[i2]
        sqr_dist += delta * delta
        point_in_box[i2] = -box_half_size[i2]
    elif point_in_box[i2] > box_half_size[i2]:
        delta = point_in_box[i2] - box_half_size[i2]
        sqr_dist += delta * delta
        point_in_box[i2] = box_half_size[i2]

    return sqr_dist, line_parameter


def _case_00(i0, i1, i2, point_in_box, direction_in_box, box_half_size):
    line_parameter = (box_half_size[i0] - point_in_box[i0]) / direction_in_box[i0]
    point_in_box[i0] = box_half_size[i0]
    box_half_size_i12 = box_half_size[[i1, i2]]
    new_point_in_box = np.clip(point_in_box[[i1, i2]], -box_half_size_i12, box_half_size_i12)
    deltas = point_in_box[[i1, i2]] - new_point_in_box
    point_in_box[[i1, i2]] = new_point_in_box
    return np.dot(deltas, deltas), line_parameter


def _case_000(point_in_box, box_half_size):
    new_point_in_box = np.clip(point_in_box, -box_half_size, box_half_size)
    deltas = point_in_box - new_point_in_box
    point_in_box[:] = new_point_in_box
    return np.dot(deltas, deltas)



import numpy as np
import pytransform3d.transformations as pt


def point_to_ellipsoid(
        point, ellipsoid2origin, radii, epsilon=1e-16, max_iter=64):
    """Compute the shortest distance between point and ellipsoid.

    Implementation adapted from 3D Game Engine Design by David H. Eberly.

    Geometric Tools, Inc.
    http://www.geometrictools.com
    Copyright (c) 1998-2006.  All Rights Reserved

    The Wild Magic Version 4 Foundation Library source code is supplied
    under the terms of the license agreement
    (http://www.geometrictools.com/License/Wm4FoundationLicense.pdf)
    and may not be copied or disclosed except in accordance with the terms
    of that agreement.

    Parameters
    ----------
    point : array, shape (3,)
        3D point.

    ellipsoid2origin : array, shape (4, 4)
        Pose of the ellipsoid.

    radii : array, shape (3,)
        Radii of the ellipsoid.

    epsilon : float, optional (default: 1e-16)
        Values smaller than epsilon are considered to be 0.

    max_iter : int, optional (default: 64)
        Maximum number of iterations of the optimization.

    Returns
    -------
    dist : float
        Shortest distance.

    closest_point_ellipsoid : array, shape (3,)
        Closest point on ellipsoid.
    """
    # compute coordinates of point in ellipsoid coordinate system
    point_in_ellipsoid = pt.transform(pt.invert_transform(ellipsoid2origin),
                                      pt.vector_to_point(point))[:3]

    radii2 = radii ** 2
    point2 = point_in_ellipsoid ** 2
    radii2point2 = radii2 * point2

    # initial guess
    if np.linalg.norm(point_in_ellipsoid / radii) < 1.0:
        # TODO in case we want to compute the distance to the surface:
        #t = 0.0  # and don't return here:
        return 0.0, point
    else:
        t = max(radii) * np.linalg.norm(point_in_ellipsoid)

    # Newton's method
    for i in range(max_iter):
        pqr = t + radii2
        pqr2 = pqr ** 2
        s = (pqr2[0] * pqr2[1] * pqr2[2]
             - radii2point2[0] * pqr2[1] * pqr2[2]
             - radii2point2[1] * pqr2[0] * pqr2[2]
             - radii2point2[2] * pqr2[0] * pqr2[1])
        if abs(s) < epsilon:
            break

        pq = pqr[0] * pqr[1]
        pr = pqr[0] * pqr[2]
        qr = pqr[1] * pqr[2]
        pqr_ = pqr[0] * pqr[1] * pqr[2]
        ds = 2.0 * (pqr_ * (qr + pr + pq)
                    - radii2point2[0] * qr * (pqr[1] + pqr[2])
                    - radii2point2[1] * pr * (pqr[0] + pqr[2])
                    - radii2point2[2] * pq * (pqr[0] + pqr[1]))
        t -= s / ds

    contact_point_in_ellipsoid = radii2 * point_in_ellipsoid / pqr
    diff = contact_point_in_ellipsoid - point_in_ellipsoid

    closest_point_ellipsoid = pt.transform(
        ellipsoid2origin, pt.vector_to_point(contact_point_in_ellipsoid))[:3]

    return np.linalg.norm(diff), closest_point_ellipsoid



"""Containment methods compute bounding volumes."""
import numpy as np
from .geometry import (
    convert_box_to_vertices, cylinder_extreme_along_direction,
    capsule_extreme_along_direction)


def axis_aligned_bounding_box(P):
    """Compute axis-aligned bounding box (AABB) that contains points.

    Parameters
    ----------
    P : array, shape (n_points, 3)
        3D points.

    Returns
    -------
    mins : array, shape (3,)
        Minimum coordinates.

    maxs : array, shape (3,)
        Maximum coordinates.
    """
    return np.min(P, axis=0), np.max(P, axis=0)


def sphere_aabb(center, radius):
    """Compute axis-aligned bounding box of sphere.

    Parameters
    ----------
    center : array, shape (3,)
        Center of the sphere.

    radius : float
        Radius of the sphere.

    Returns
    -------
    mins : array, shape (3,)
        Minimum coordinates.

    maxs : array, shape (3,)
        Maximum coordinates.
    """
    return center - radius, center + radius


def box_aabb(box2origin, size):
    """Compute axis-aligned bounding box of an oriented box.

    Parameters
    ----------
    box2origin : array, shape (4, 4)
        Pose of the box.

    size : array, shape (3,)
        Size of the box.

    Returns
    -------
    mins : array, shape (3,)
        Minimum coordinates.

    maxs : array, shape (3,)
        Maximum coordinates.
    """
    vertices = convert_box_to_vertices(box2origin, size)
    return axis_aligned_bounding_box(vertices)


XM = np.array([-1.0, 0.0, 0.0])
YM = np.array([0.0, -1.0, 0.0])
ZM = np.array([0.0, 0.0, -1.0])
XP = np.array([1.0, 0.0, 0.0])
YP = np.array([0.0, 1.0, 0.0])
ZP = np.array([0.0, 0.0, 1.0])


def cylinder_aabb(cylinder2origin, radius, length):
    """Compute axis-aligned bounding box of cylinder.

    Parameters
    ----------
    cylinder2origin : array, shape (4, 4)
        Pose of the cylinder.

    radius : float
        Radius of the cylinder.

    length : float
        Length of the cylinder.

    Returns
    -------
    mins : array, shape (3,)
        Minimum coordinates.

    maxs : array, shape (3,)
        Maximum coordinates.
    """
    negative_vertices = np.vstack((
        cylinder_extreme_along_direction(XM, cylinder2origin, radius, length),
        cylinder_extreme_along_direction(YM, cylinder2origin, radius, length),
        cylinder_extreme_along_direction(ZM, cylinder2origin, radius, length),
    ))
    mins = np.min(negative_vertices, axis=0)
    positive_vertices = np.vstack((
        cylinder_extreme_along_direction(XP, cylinder2origin, radius, length),
        cylinder_extreme_along_direction(YP, cylinder2origin, radius, length),
        cylinder_extreme_along_direction(ZP, cylinder2origin, radius, length),
    ))
    maxs = np.max(positive_vertices, axis=0)
    return mins, maxs


def capsule_aabb(capsule2origin, radius, height):
    """Compute axis-aligned bounding box of a capsule.

    Parameters
    ----------
    capsule2origin : array, shape (4, 4)
        Pose of the capsule.

    radius : float
        Radius of the capsule.

    height : float
        Height of the capsule.

    Returns
    -------
    mins : array, shape (3,)
        Minimum coordinates.

    maxs : array, shape (3,)
        Maximum coordinates.
    """
    negative_vertices = np.vstack((
        capsule_extreme_along_direction(XM, capsule2origin, radius, height),
        capsule_extreme_along_direction(YM, capsule2origin, radius, height),
        capsule_extreme_along_direction(ZM, capsule2origin, radius, height),
    ))
    mins = np.min(negative_vertices, axis=0)
    positive_vertices = np.vstack((
        capsule_extreme_along_direction(XP, capsule2origin, radius, height),
        capsule_extreme_along_direction(YP, capsule2origin, radius, height),
        capsule_extreme_along_direction(ZP, capsule2origin, radius, height),
    ))
    maxs = np.max(positive_vertices, axis=0)
    return mins, maxs



import numpy as np
from ..geometry import hesse_normal_form, convert_segment_to_line


def point_to_plane(point, plane_point, plane_normal, signed=False):
    """Compute the shortest distance between a point and a plane.

    Parameters
    ----------
    point : array, shape (3,)
        3D point.

    plane_point : array, shape (3,)
        Point on the plane.

    plane_normal : array, shape (3,)
        Normal of the plane. We assume unit length.

    signed : bool, optional (default: False)
        Should the distance have a sign?

    Returns
    -------
    dist : float
        The shortest distance between point and plane. A sign indicates
        the direction along the normal.

    closest_point_plane : array, shape (3,)
        Closest point on plane.
    """
    t = np.dot(plane_normal, point - plane_point)
    closest_point_plane = point - t * plane_normal
    if not signed:
        t = abs(t)
    return t, closest_point_plane


def line_to_plane(
        line_point, line_direction, plane_point, plane_normal, epsilon=1e-6):
    """Compute the shortest distance from line to plane.

    Parameters
    ----------
    line_point : array, shape (3,)
        Point on line.

    line_direction : array, shape (3,)
        Direction of the line. This is assumed to be of unit length.

    plane_point : array, shape (3,)
        Point on the plane.

    plane_normal : array, shape (3,)
        Normal of the plane. We assume unit length.

    epsilon : float, optional (default: 1e-6)
        Values smaller than epsilon are considered to be 0.

    Returns
    -------
    dist : float
        The shortest distance between line and plane.

    closest_point_line : array, shape (3,)
        Closest point on line.

    closest_point_plane : array, shape (3,)
        Closest point on plane.
    """
    intersection, t = _line_to_plane(
        line_point, line_direction, plane_point, plane_normal, epsilon)

    if intersection:
        closest_point_line = line_point + t * line_direction
        return 0.0, closest_point_line, closest_point_line

    dist, closest_point_plane = point_to_plane(
        line_point, plane_point, plane_normal)
    return dist, line_point, closest_point_plane


def _line_to_plane(line_point, line_direction, plane_point, plane_normal,
                   epsilon=1e-6):
    length_of_dir_projected_on_normal = np.dot(line_direction, plane_normal)
    if length_of_dir_projected_on_normal * length_of_dir_projected_on_normal < epsilon:
        # line parallel to plane
        return False, 0.0

    _, d = hesse_normal_form(plane_point, plane_normal)
    t = (d - np.dot(plane_normal, line_point)) / np.dot(plane_normal, line_direction)
    return True, t


def line_segment_to_plane(
        segment_start, segment_end, plane_point, plane_normal, epsilon=1e-6):
    """Compute the shortest distance between line segment and plane.

    Parameters
    ----------
    segment_start : array, shape (3,)
        Start point of segment.

    segment_end : array, shape (3,)
        End point of segment.

    plane_point : array, shape (3,)
        Point on the plane.

    plane_normal : array, shape (3,)
        Normal of the plane. We assume unit length.

    epsilon : float, optional (default: 1e-6)
        Values smaller than epsilon are considered to be 0.

    Returns
    -------
    dist : float
        The shortest distance between line segment and plane.

    closest_point_segment : array, shape (3,)
        Closest point on line segment.

    closest_point_plane : array, shape (3,)
        Closest point on plane.
    """
    segment_direction, segment_length = convert_segment_to_line(
        segment_start, segment_end)
    line_intersects, t = _line_to_plane(
        segment_start, segment_direction, plane_point, plane_normal, epsilon)

    if line_intersects:
        if 0 <= t <= segment_length:
            closest_point_segment = segment_start + t * segment_direction
            return 0.0, closest_point_segment, closest_point_segment

        if t < 0.0:
            closest_point_segment = segment_start
        else:
            closest_point_segment = segment_end
    else:
        closest_point_segment = segment_start

    dist, closest_point_plane = point_to_plane(
        closest_point_segment, plane_point, plane_normal)
    return dist, closest_point_segment, closest_point_plane



from . import gjk


def detect(bvh, collision_margin=1e-3):
    """Detect self collisions of a robot represented by a BVH.

    Parameters
    ----------
    bvh : BoundingVolumeHierarchy
        Bounding volume hierarchy that contains colliders of a robot. Note that
        the attribute bvh.self_collision_whitelists_ has to be filled.
        Otherwise a collection of a collider with itself or direct neighbors
        will be considered a self collision!

    collision_margin : float, optional (default: 0.001)
        Distance between colliders that is considered to be a collision.

    Returns
    -------
    contacts : dict
        Maps each collider frame to a boolean indicating whether it is in
        contact with another collider or not.
    """
    contacts = {}
    for frame, collider in bvh.colliders_.items():
        candidates = bvh.aabb_overlapping_colliders(
            collider, whitelist=bvh.self_collision_whitelists_[frame])
        contacts[frame] = False
        for frame2, collider2 in candidates.items():
            dist, _, _, _ = gjk.gjk_with_simplex(collider, collider2)
            if dist < collision_margin:
                contacts[frame] = True
                break
    return contacts
